#!groovy
import org.jfrog.stork.model.ModuleConfigContext
import org.jfrog.stork.model.module.JFrogModule

@Library('stork@infra/DEVF-858') _

static def sonarJobForBranch(String sourceBranch) {
    'Artifactory/artifactory-sonar-pipe/' + URLEncoder.encode(sourceBranch, "UTF-8")
}

storkConfig { ModuleConfigContext modCtx ->

    echo "Setting 'artifactory-oss' config"
    modCtx.moduleName = 'artifactory-oss'

    modCtx.init = { JFrogModule module ->
        module.tool = 'mvn-3.6.3'
        module.allowVersionOverride('maven.plugin.surefire.version', 'RTFACT-15954')
        module.deployDocker = true
        module.dockerImageNames = ['jfrog/artifactory-oss']
        module.dockerRegistry = 'art.jfrog.io'
        module.dockerLocalRepo = 'art-docker-dev-local'
        module.setJDKName('jdk-11')

        module.maxBuilds = 50
    }

    modCtx.preBuild = { JFrogModule module ->
        echo "${module.name} extension env setup"
        module.settings.mailRecipients = module.settings.isRelease() ? 'stork@jfrog.com' : null
        module.mvnProfiles = 'release,deb,rpm,docker'
        module.addMavenOption("-DskipTests -T 1C -Dorg.slf4j.simpleLogger.showDateTime=true")

        def artifactoryVersion = module.getDockerTagVersion()
        env.TARGET_VERSION = artifactoryVersion
        env.DOCKER_COMPOSE_REGISTRY = module.settings.isRelease() ? "docker.bintray.io" : "docker.jfrog.io"

        echo "Artifactory Tag Version is ${artifactoryVersion}"

        if (!module.settings.isRelease()) {
            //Added this since we see some cases where classes are left in the target and mess up the build and we need them cleaned.
//            boolean cleanMvn = env.CLEAN_MVN ? env.CLEAN_MVN.toBoolean() : false
//            if (!cleanMvn) {
                //Default mvnCommand for the job is clean install
                module.setMavenPhases("clean install")
                module.addMavenOption("-Pdist-clean,reports-clean")
//            }
        }

        withCredentials([string(credentialsId: 'rpm-sign-passphrase', variable: 'SECRET')]) {
            wrap([$class: 'MaskPasswordsBuildWrapper', varPasswordPairs: [[password: SECRET, var: 'SECRET']]]) {
                module.settings.envVars << "passphrase=${SECRET}".toString()
            }
        }

        storkCommon.setUiBuildResource(module.settings)

        if (module.settings.isRelease() && !module.downstream) {
            // STRK-4 Update ArtifactoryVersion.java
            storkRelease.updateArtifactoryVersion(module)
        }

        // DEVF-684 - Disable automatic trigger of sonar due to jenkins issues
        /*Map<String, Closure> preTasks = [:]
        pushSonarJob(1, module, false, preTasks)
        parallel preTasks*/
    }

    modCtx.preTransformVersion = { JFrogModule module ->
        echo "${module.name} preTransformVersion"
        if (params.UPDATE_FRONTEND) {
            def version = params.FRONTEND_VERSION != "" ? params.FRONTEND_VERSION : getFrontendLatestVersion()
            echo "frontend version - $version"
            storkRelease.setFrontendVersionParamForRelease(version)
        }
    }

    modCtx.endOfBuild = { JFrogModule module ->
        if (module.settings.isRelease()) {
            echo "In release mode, no post build actions!"
            return
        }
        if (!module.deployArtifacts) {
            echo "Artifacts were not deployed to Artifactory, no post build actions!"
            return
        }
        echo "Post Build for ${module.name}"
        String ARTIFACTORY_TAG_VERSION = module.getDockerTagVersion()

        if (!module.isSuccessfulBuild()) {
            echo "Build was not successful, skipping tests!"
            return
        }

        Map<String, Closure> postTasks = [:]
        if (module.settings.isMultibranch() && !module.isMaster()) {
            // Just run sanity without waiting
            postTasks['Trigger artifactory-sanity-5.x-branches'] = {
                build job: "artifactory/test/artifactory-sanity/master",
                        parameters: [
                                string(name: 'artifactory_version', value: ARTIFACTORY_TAG_VERSION),
                                booleanParam(name: 'oss', value: true),
                                booleanParam(name: 'pro', value: false),
                                booleanParam(name: 'ha', value: false),
                                booleanParam(name: 'conan', value: false),
                                string(name: 'test_scope', value: 'extended')
                        ], propagate: false, wait: false
            }
        } else {
            // In non multibranch run strict tests if set
            boolean strict = env.STRICT ? env.STRICT.toBoolean() : false
            int nRetry = strict ? 2 : 1

            postTasks['Trigger artifactory-sanity-5.x'] = {
                retry(nRetry) {
                    build job: "artifactory/test/artifactory-sanity/master",
                            parameters: [
                                    string(name: 'artifactory_version', value: ARTIFACTORY_TAG_VERSION),
                                    booleanParam(name: 'oss', value: true),
                                    booleanParam(name: 'pro', value: false),
                                    booleanParam(name: 'ha', value: false),
                                    booleanParam(name: 'conan', value: false),
                                    string(name: 'test_scope', value: 'extended')
                            ], propagate: strict, wait: strict

                }
            }
            postTasks['artifactory-oss-5.x-db-unitest'] = {
                retry(nRetry) {
                    build job: 'Artifactory/artifactory-oss-5.x-db-unitest',
                            parameters: [
                                    booleanParam(name: 'RUN_TEST_DERBY', value: true),
                                    booleanParam(name: 'RUN_TEST_MYSQL', value: true),
                                    booleanParam(name: 'RUN_TEST_MSSQL', value: false), //mssql runs in its own job on windows os
                                    booleanParam(name: 'RUN_TEST_ORACLE', value: true),
                                    booleanParam(name: 'RUN_TEST_POSTGRESQL', value: true),
                                    booleanParam(name: 'RUN_TEST_MARIADB', value: true),
                                    stringParam(name: 'MVN_OVERWRITE_CMD',
                                            value: '-U -f /workspace/artifactory-oss/storage/db/pom.xml -Prelease clean test -fae ' +
                                                    '-Dmaven.repo.local=/workspace/artifactory-pro/docker-builder/build/.repository ' +
                                                    '-Dmaven.test.failure.ignore=true  -s ' +
                                                    '/workspace/artifactory-pro/docker-builder/build/jenkins-settings.xml')
                            ], propagate: strict, wait: strict
                }
            }
        }
        //parallel postTasks
    }
}

private void pushSonarJob(int nRetry, JFrogModule module, boolean strict, Map<String, Closure> postTasks) {
    postTasks['Trigger Artifactory Sonar'] = {
        retry(nRetry) {
            build job: sonarJobForBranch(module.getSourceBranch() as String),
                    parameters: [
                            stringParam(name: 'BRANCH_NAME', value: module.getSourceBranch()),
                            stringParam(name: 'REPOSITORY', value: ''),
                            stringParam(name: 'COMMIT', value: '')
                    ], propagate: strict, wait: strict
        }
    }
}

def getFrontendLatestVersion() {
    withCredentials(
            [[$class          : 'UsernamePasswordMultiBinding', credentialsId: 'f0819c4f-7229-4676-b391-77fcb19f95e4',
              passwordVariable: 'artPassword', usernameVariable: 'artUser']]) {
        def frontendVersionAQL = "items.find({\"\$and\":[{\"repo\":\"libs-releases-local\",\"type\":\"folder\"," +
                "\"path\":\"com/jfrog/platform/jfrog-frontend-bin\"}]}).sort({\"\$desc\":[\"created\"]}).limit(1)"

        String version = sh returnStdout: true,
                script: "curl -u $artUser:$artPassword --fail --silent --show-error -H \'Content-Type: " +
                        "text/plain\' https://entplus.jfrog.io/artifactory/api/search/aql -d \'$frontendVersionAQL\' | jq \'.results[].name\'"
        return version.replaceAll("\"", "").trim()
    }
}
