/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2020 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.artifactory.metrics.framework.providers;

import org.artifactory.storage.binstore.service.GarbageCollectionDatastoreService;
import org.artifactory.storage.binstore.service.GarbageCollectorInfo;
import org.artifactory.storage.binstore.service.GarbageCollectorStrategy;
import org.jfrog.metrics.model.Metric;
import org.jfrog.metrics.model.MetricProvider;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.TimeUnit;

/**
 * @author leela
 */
public class GarbageCollectionMetricProvider implements MetricProvider {
    private static final Logger log = LoggerFactory.getLogger(GarbageCollectionMetricProvider.class);

    private long startDate;

    private GarbageCollectionDatastoreService garbageCollectionDatastoreService;

    public GarbageCollectionMetricProvider(GarbageCollectionDatastoreService gcService) {
        this.garbageCollectionDatastoreService = gcService;
    }

    @Override
    public String getName() {
        return "jfrt_artifacts_gc";
    }

    @Override
    public String getDescription() {
        return "Artifactory Garbage Collection Metrics Provider";
    }

    @Override
    public long getRefreshIntervalMillis() {
        return TimeUnit.SECONDS.toMillis(60);
    }

    @Override
    public List<Metric> collect() {
        List<Metric> metrics = new ArrayList<>();
        GarbageCollectorInfo data = garbageCollectionDatastoreService.getLastGCRunInfo();
        if (null == data) {
            log.debug("GC Store data is null, hence ignoring to collect metrics.");
            return metrics;
        }
        startDate = data.getGcStartTime();
        float durationInSecs = ((float) (data.gcEndTime - data.getGcStartTime())) / 1000;
        Map<String, Object> labels = new HashMap<>();
        labels.put("start", startDate);
        labels.put("end", data.gcEndTime);
        labels.put("status", data.status.name());
        labels.put("type", data.strategy.name());


        Metric durationMetric = getMetric(labels, "duration", Metric.MetricType.SECONDS,
                "Time taken for garbage collection", durationInSecs);
        Metric sizeCleanedMetric = getMetric(labels, "size_cleaned", Metric.MetricType.BYTES,
                "Total Bytes recovered by Garbage Collection", data.totalSizeCleaned.get());
        Metric totalBinariesMetric = getMetric(labels, "binaries", Metric.MetricType.COUNTER,
                "Total number of binaries removed by Garbage Collection", data.binariesCleaned.get());
        metrics.add(durationMetric);
        metrics.add(totalBinariesMetric);
        metrics.add(sizeCleanedMetric);
        if (data.strategy.equals(GarbageCollectorStrategy.FULL)) {
            long currentSize = data.initialSize - data.totalSizeCleaned.get();
            Metric currentSizeMetric = getMetric(labels, "current_size", Metric.MetricType.BYTES,
                    "Total space occupied by binaries after Garbage Collection", currentSize);
            metrics.add(currentSizeMetric);
        }

        return metrics;
    }

    private Metric getMetric(Map<String, Object> labels, String metricName, Metric.MetricType metricType,
            String description, float metricValue) {
        return Metric.builder()
                .name(metricName)
                .type(metricType)
                .labels(labels)
                .description(description)
                .value(metricValue)
                .build();
    }

    @Override
    public ReportType getReportType() {
        return ReportType.DEDUPED;
    }

    @Override
    public long getLastUpdatedMillis() {
        if (null != garbageCollectionDatastoreService.getLastGCRunInfo()) {
            startDate = garbageCollectionDatastoreService.getLastGCRunInfo().getGcStartTime();
        }
        return startDate;
    }
}
