package org.artifactory.traffic;

import ch.qos.logback.core.rolling.RollingFileAppender;

import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

public class TrafficTimeBasedRollingFileAppender extends RollingFileAppender {

    private static final String ROLLOVER_THREAD_ENABLED_PROPERTY = "artifactory.traffic.rollover.thread.enabled";

    private ScheduledExecutorService executorService;
    private Long intervalMs;

    public Long getIntervalMs() {
        return intervalMs;
    }

    public void setIntervalMs(Long intervalMs) {
        this.intervalMs = intervalMs;
    }

    @Override
    public void start() {
        super.start();
        if (isRolloverEnabled()) {
            long initialDelay = calcInitialDelay();
            scheduleRollover(initialDelay);
        }
    }

    @Override
    public void stop() {
        if (isRolloverEnabled()) {
            shutdownExecutor();
        }
        super.stop();
    }

    protected void scheduleRollover(long initialDelay) {
        executorService = Executors.newSingleThreadScheduledExecutor(DaemonThread::new);
        executorService.scheduleAtFixedRate(this::internalRollover, initialDelay, intervalMs, TimeUnit.MILLISECONDS);
    }

    protected long calcLifetime() {
        return System.currentTimeMillis() - getContext().getBirthTime();
    }

    protected void shutdownExecutor() {
        try {
            executorService.shutdownNow();
            executorService.awaitTermination(1000, TimeUnit.MILLISECONDS);
        } catch (Exception e) {
            addWarn("Error while executor shutdown", e);
        }
    }

    protected void internalRollover() {
        try {
            rollover();
            addInfo("Traffic logger rollover completed ");
        } catch (Exception e) {
            addError("Failed to rollover traffic log", e);
        }
    }

    protected String getStartTrafficRolloverSystemProperty() {
        return System.getProperty(ROLLOVER_THREAD_ENABLED_PROPERTY, "false");
    }

    private boolean isRolloverEnabled() {
        String startTrafficRolloverThread = getStartTrafficRolloverSystemProperty();
        boolean isRolloverThreadEnabled = Boolean.parseBoolean(startTrafficRolloverThread);
        return isRolloverThreadEnabled && intervalMs != null;
    }

    private long calcInitialDelay() {
        long lifetime = calcLifetime();
        return intervalMs - (lifetime % intervalMs);
    }

    private static class DaemonThread extends Thread {

        public DaemonThread(Runnable runnable) {
            super(runnable);
            setName("appender-rollover-thread");
            setDaemon(true);
        }
    }

}
