package org.artifactory.schedule;

import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.spring.ArtifactoryApplicationContext;
import org.artifactory.storage.db.locks.service.DbLocksService;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.jfrog.common.concurrency.ConflictsGuard;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;
import org.mockito.MockitoSession;
import org.mockito.quality.Strictness;
import org.quartz.JobExecutionContext;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.util.Collections;

import static java.util.concurrent.TimeUnit.SECONDS;
import static org.artifactory.concurrent.NonStandardLockCategory.CLUSTER_SINGLETON;
import static org.junit.Assert.assertFalse;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoInteractions;
import static org.mockito.Mockito.verifyNoMoreInteractions;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertTrue;


@Test
public class HaQuartzTaskTest extends ArtifactoryHomeBoundTest {

    @Mock
    private ConflictsGuard<String> conflictsGuard;
    @Mock
    private DbLocksService dbLocksService;
    @Mock
    private QuartzTask quartzTask;
    @Mock
    private ArtifactoryApplicationContext artifactoryContext;
    private String stopMarkerLockPrefix = "stop_marker_" + Job1.getMyJobCommand() + "_";
    private String pauseMarkerLockPrefix = "pause_marker_" + Job1.getMyJobCommand() + "_";
    private MockitoSession mockitoSession;


    @BeforeMethod
    public void init() {
        mockitoSession = Mockito.mockitoSession()
                .strictness(Strictness.LENIENT)
                .initMocks(this)
                .startMocking();

        ArtifactoryContextThreadBinder.bind(artifactoryContext);
        MockitoAnnotations.initMocks(this);
    }

    @AfterMethod
    public void tearDown() {
        mockitoSession.finishMocking();
    }

    @Test
    public void testPauseWithTaskTokenNull() {
        defineMockJobType();
        HaQuartzTask haQuartzTask = new HaQuartzTask(quartzTask, conflictsGuard, dbLocksService);
        haQuartzTask.pause(true, null);
        verify(quartzTask).pause(true);
        verifyNoMoreInteractions(conflictsGuard);
    }

    @Test
    public void testPauseWithTaskTokenAndLockNotAcquired() throws InterruptedException {
        defineMockJobType();
        String pausingTaskToken = "pausingTaskToken";
        String uniqueMarkerLock = pauseMarkerLockPrefix + pausingTaskToken;
        when(conflictsGuard.isLocked(uniqueMarkerLock)).thenReturn(false);
        when(conflictsGuard.tryToLock(uniqueMarkerLock, 0, SECONDS)).thenReturn(true);
        HaQuartzTask haQuartzTask = new HaQuartzTask(quartzTask, conflictsGuard, dbLocksService);

        haQuartzTask.pause(true, pausingTaskToken);

        verify(quartzTask).pause(true);
        verify(conflictsGuard).tryToLock(uniqueMarkerLock, 0, SECONDS);
    }

    @Test
    public void testPauseWithTaskTokenAndLockAlreadyAcquired() throws InterruptedException {
        defineMockJobType();
        String pausingTaskToken = "pausingTaskToken";
        String uniqueMarkerLock = pauseMarkerLockPrefix + pausingTaskToken;
        when(conflictsGuard.isLocked(uniqueMarkerLock)).thenReturn(true);
        when(conflictsGuard.tryToLock(uniqueMarkerLock, 0, SECONDS)).thenReturn(true);
        HaQuartzTask haQuartzTask = new HaQuartzTask(quartzTask, conflictsGuard, dbLocksService);

        haQuartzTask.pause(true, pausingTaskToken);

        verify(quartzTask).pause(true);
        verify(conflictsGuard, times(1)).isLocked(uniqueMarkerLock);
        verify(conflictsGuard, times(0)).tryToLock(uniqueMarkerLock, 0, SECONDS);
    }

    @Test
    public void testStopWithTaskTokenNull() {
        defineMockJobType();
        HaQuartzTask haQuartzTask = new HaQuartzTask(quartzTask, conflictsGuard, dbLocksService);
        haQuartzTask.stop(true, null);
        verify(quartzTask).stop(true);
        verifyNoMoreInteractions(conflictsGuard);
    }

    @Test
    public void testStopWithTaskTokenAndLockNotAcquired() throws InterruptedException {
        defineMockJobType();
        String stopTaskToken = "stopTaskToken";
        String uniqueMarkerLock = stopMarkerLockPrefix + stopTaskToken;
        when(conflictsGuard.isLocked(uniqueMarkerLock)).thenReturn(false);
        when(conflictsGuard.tryToLock(uniqueMarkerLock, 0, SECONDS)).thenReturn(true);
        HaQuartzTask haQuartzTask = new HaQuartzTask(quartzTask, conflictsGuard, dbLocksService);

        haQuartzTask.stop(true, stopTaskToken);

        verify(quartzTask).stop(true);
        assertTrue(conflictsGuard.tryToLock(uniqueMarkerLock, 0, SECONDS));
    }

    @Test
    public void testStopWithTaskTokenAndLockAlreadyAcquired() throws InterruptedException {
        defineMockJobType();
        String stopTaskToken = "stopTaskToken";
        String uniqueMarkerLock = stopMarkerLockPrefix + stopTaskToken;
        when(conflictsGuard.isLocked(uniqueMarkerLock)).thenReturn(true);
        when(conflictsGuard.tryToLock(uniqueMarkerLock, 0, SECONDS)).thenReturn(true);
        HaQuartzTask haQuartzTask = new HaQuartzTask(quartzTask, conflictsGuard, dbLocksService);

        haQuartzTask.stop(true, stopTaskToken);

        verify(quartzTask).stop(true);
        verify(conflictsGuard, times(1)).isLocked(uniqueMarkerLock);
        verify(conflictsGuard, times(0)).tryToLock(uniqueMarkerLock, 0, SECONDS);
    }

    @Test
    public void testResumeTaskWhenTokenIsNull() {
        defineMockJobType();
        when(quartzTask.resume()).thenReturn(true);
        HaQuartzTask haQuartzTask = new HaQuartzTask(quartzTask, conflictsGuard, dbLocksService);
        assertTrue(haQuartzTask.resume(null));
        verifyNoInteractions(conflictsGuard);
    }

    @Test
    public void testResumePausedTask() {
        defineMockJobType();
        String pausingTaskToken = "pausingTaskToken";
        String uniqueMarkerLock = pauseMarkerLockPrefix + pausingTaskToken;
        when(quartzTask.resume()).thenReturn(true);
        when(conflictsGuard.isLocked(uniqueMarkerLock)).thenReturn(false);
        HaQuartzTask haQuartzTask = new HaQuartzTask(quartzTask, conflictsGuard, dbLocksService);
        haQuartzTask.pause(true, pausingTaskToken);

        boolean isScheduled = haQuartzTask.resume(pausingTaskToken);
        assertTrue(isScheduled);
        verify(conflictsGuard).unlock(uniqueMarkerLock);
    }

    @Test
    public void testResumeStoppedTask() {
        defineMockJobType();
        String stoppingTaskToken = "stoppingTaskToken";
        String uniqueMarkerLock = stopMarkerLockPrefix + stoppingTaskToken;
        when(quartzTask.resume()).thenReturn(true);
        when(conflictsGuard.isLocked(uniqueMarkerLock)).thenReturn(false);
        HaQuartzTask haQuartzTask = new HaQuartzTask(quartzTask, conflictsGuard, dbLocksService);
        haQuartzTask.stop(true, stoppingTaskToken);
        boolean isScheduled = haQuartzTask.resume(stoppingTaskToken);
        assertTrue(isScheduled);
        verify(conflictsGuard).unlock(uniqueMarkerLock);
    }

    @Test
    public void testResumeNoStoppedTask() {
        defineMockJobType();
        String stoppingTaskToken = "stoppingTaskToken";
        String uniqueMarkerLock = pauseMarkerLockPrefix + stoppingTaskToken;
        when(quartzTask.resume()).thenReturn(true);
        when(conflictsGuard.isLocked(uniqueMarkerLock)).thenReturn(false);
        HaQuartzTask haQuartzTask = new HaQuartzTask(quartzTask, conflictsGuard, dbLocksService);
        boolean isScheduled = haQuartzTask.resume(stoppingTaskToken);
        assertTrue(isScheduled);
        verify(conflictsGuard, times(0)).unlock(uniqueMarkerLock);
    }

    @Test
    public void testStartedWhenTaskIsRunningAndNoStopMarker() {
        defineMockJobType();
        when(quartzTask.started()).thenReturn(true);
        when(quartzTask.getToken()).thenReturn("dummyToken");
        when(dbLocksService.isLockedByPrefixExcluding(stopMarkerLockPrefix, stopMarkerLockPrefix + "dummyToken", false))
                .thenReturn(false);
        HaQuartzTask haQuartzTask = new HaQuartzTask(quartzTask, conflictsGuard, dbLocksService);
        assertTrue(haQuartzTask.started());
    }

    @Test
    public void testStartedWhenTaskIsRunningAndStopMarkerExists() {
        defineMockJobType();
        when(quartzTask.started()).thenReturn(true);
        when(quartzTask.getToken()).thenReturn("dummyToken");
        when(dbLocksService.isLockedByPrefixExcluding(stopMarkerLockPrefix, stopMarkerLockPrefix + "dummyToken", false))
                .thenReturn(true);
        HaQuartzTask haQuartzTask = new HaQuartzTask(quartzTask, conflictsGuard, dbLocksService);
        assertFalse(haQuartzTask.started());
    }

    @Test
    public void testStartedWhenTaskIsNotRunningAndNoStopMarker() {
        defineMockJobType();
        when(quartzTask.started()).thenReturn(false);
        when(dbLocksService.isLocked(CLUSTER_SINGLETON.getName(), quartzTask.getTaskLockKeyIncludingKeyAttributes(
                Collections.emptyList()))).thenReturn(false);
        HaQuartzTask haQuartzTask = new HaQuartzTask(quartzTask, conflictsGuard, dbLocksService);
        assertFalse(haQuartzTask.started());
    }

    @Test
    public void testStartedWhenTaskIsNotRunningAndStopMarkerExists() {
        defineMockJobType();
        when(quartzTask.started()).thenReturn(false);
        when(dbLocksService.isLockedByPrefixExcluding(stopMarkerLockPrefix, stopMarkerLockPrefix + "dummyToken", false))
                .thenReturn(true);
        HaQuartzTask haQuartzTask = new HaQuartzTask(quartzTask, conflictsGuard, dbLocksService);
        assertFalse(haQuartzTask.started());
    }

    @Test
    public void testIsRunningOnClusterWhenJobRunningOnCurrentNode() {
        when(quartzTask.isRunning()).thenReturn(true);
        defineMockJobType();
        HaQuartzTask haQuartzTask = new HaQuartzTask(quartzTask, conflictsGuard, dbLocksService);
        assertTrue(haQuartzTask.isRunningOnCluster());
    }

    @Test
    public void testIsRunningOnClusterWhenJobIsRunningOnDifferentNode() {
        defineMockJobType();
        when(quartzTask.isRunning()).thenReturn(false);
        when(dbLocksService.isLocked(CLUSTER_SINGLETON.getName(), quartzTask.getTaskLockKeyIncludingKeyAttributes(
                Collections.emptyList()))).thenReturn(true);
        HaQuartzTask haQuartzTask = new HaQuartzTask(quartzTask, conflictsGuard, dbLocksService);
        assertTrue(haQuartzTask.isRunningOnCluster());
    }

    @Test
    public void testIsRunningOnClusterWhenJobIsNotRunning() {
        defineMockJobType();
        when(quartzTask.isRunning()).thenReturn(false);
        when(dbLocksService.isLocked(CLUSTER_SINGLETON.getName(), quartzTask.getTaskLockKeyIncludingKeyAttributes(
                Collections.emptyList()))).thenReturn(false);
        HaQuartzTask haQuartzTask = new HaQuartzTask(quartzTask, conflictsGuard, dbLocksService);
        assertFalse(haQuartzTask.isRunningOnCluster());
    }

    @Test
    public void testProcessActiveOnClusterWhenProcessIsActiveOnCurrentNode() {
        when(quartzTask.processActive()).thenReturn(true);
        defineMockJobType();
        HaQuartzTask haQuartzTask = new HaQuartzTask(quartzTask, conflictsGuard, dbLocksService);
        assertTrue(haQuartzTask.processActiveOnCluster());
    }

    @Test
    public void testProcessActiveOnClusterWhenProcessIsActiveOnDifferentNode() {
        when(quartzTask.processActive()).thenReturn(false);
        when(quartzTask.isRunning()).thenReturn(true);
        defineMockJobType();
        HaQuartzTask haQuartzTask = new HaQuartzTask(quartzTask, conflictsGuard, dbLocksService);
        assertTrue(haQuartzTask.processActiveOnCluster());
    }

    @Test
    public void testProcessActiveOnClusterWhenProcessAcquiredLock() {
        when(quartzTask.processActive()).thenReturn(false);
        when(quartzTask.isRunning()).thenReturn(false);
        when(dbLocksService.isLocked(CLUSTER_SINGLETON.getName(), quartzTask.getTaskLockKeyIncludingKeyAttributes(
                Collections.emptyList()))).thenReturn(true);
        defineMockJobType();
        HaQuartzTask haQuartzTask = new HaQuartzTask(quartzTask, conflictsGuard, dbLocksService);
        assertTrue(haQuartzTask.processActiveOnCluster());
    }

    @Test
    public void testProcessActiveOnClusterWhenProcessNotActiveAndDidNotAcquiredLock() {
        defineMockJobType();
        when(quartzTask.processActive()).thenReturn(false);
        when(quartzTask.isRunning()).thenReturn(false);
        when(dbLocksService.isLocked(CLUSTER_SINGLETON.getName(), quartzTask.getTaskLockKeyIncludingKeyAttributes(
                Collections.emptyList()))).thenReturn(false);
        HaQuartzTask haQuartzTask = new HaQuartzTask(quartzTask, conflictsGuard, dbLocksService);
        assertFalse(haQuartzTask.processActiveOnCluster());
    }

    @Test
    public void testCancelTaskWithWhenWaitForRunningIsTrue() {
        defineMockJobType();
        HaQuartzTask haQuartzTask = new HaQuartzTask(quartzTask, conflictsGuard, dbLocksService);
        haQuartzTask.cancel(true, 10);
        verify(quartzTask).cancel(true, 10);
    }

    @Test
    public void testCancelTaskWithWhenWaitForRunningIsFalse() {
        defineMockJobType();
        HaQuartzTask haQuartzTask = new HaQuartzTask(quartzTask, conflictsGuard, dbLocksService);
        haQuartzTask.cancel(false, 10);
        verify(quartzTask).cancel(false, 10);
    }

    @Test
    public void testBlockIfPausedAndShouldBreakWhenJobIsPausedWithoutMarker() {
        defineMockJobType();
        when(quartzTask.blockIfPausedAndShouldBreak()).thenReturn(true);
        HaQuartzTask haQuartzTask = new HaQuartzTask(quartzTask, conflictsGuard, dbLocksService);
        assertTrue(haQuartzTask.blockIfPausedAndShouldBreak());
    }

    @Test
    public void testWaitForCompletion() {
        defineMockJobType();
        when(quartzTask.waitForCompletion(10000)).thenReturn(true);
        HaQuartzTask haQuartzTask = new HaQuartzTask(quartzTask, conflictsGuard, dbLocksService);
        haQuartzTask.waitForCompletion(10000);
        assertTrue(haQuartzTask.waitForCompletion(10000));
    }

    @SuppressWarnings("ResultOfMethodCallIgnored")
    private void defineMockJobType() {
        doReturn(Job1.class).when(quartzTask).getType();
    }

    @JobCommand(jobName = JobName.IMPORT, schedulerUser = TaskUser.SYSTEM, manualUser = TaskUser.SYSTEM)
    public static class Job1 extends QuartzCommand {

        @Override
        protected void onExecute(JobExecutionContext callbackContext) {
        }

        static JobName getMyJobCommand() {
            return Job1.class.getAnnotation(JobCommand.class).jobName();
        }
    }
}