/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.lifecycle.storage.db.init;

import lombok.SneakyThrows;
import org.artifactory.common.ha.HaNodeProperties;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.concurrent.NonStandardLockCategory;
import org.artifactory.storage.db.dao.locks.LockInfo;
import org.artifactory.storage.db.locks.service.DbLocksService;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.Optional;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.function.Supplier;

/**
 * @author Uriah Levy
 * @author Mor Merhav
 */
public class HaInitLock {
    private static final Logger log = LoggerFactory.getLogger(HaInitLock.class);
    public static final String ARTIFACTORY_INIT_LOCK_KEY = "artifactory-init";

    private final DbLocksService dbLocksService;
    private final ArtifactoryHome artifactoryHome;

    public HaInitLock(DbLocksService dbLocksService, ArtifactoryHome artifactoryHome) {
        this.dbLocksService = dbLocksService;
        this.artifactoryHome = artifactoryHome;
    }

    public void releaseMyInitLock() {
        getHaInitLockInfo().ifPresent(lockInfo -> {
            HaNodeProperties haNodeProperties = artifactoryHome.getHaNodeProperties();
            String serverId = haNodeProperties.getServerId();
            if (serverId.equals(lockInfo.getOwner())) {
                log.info("Found an init lock by current node: {}. Removing the lock.", lockInfo.getOwner());
                // We force unlock because there is no guarantee that the thread who acquired the lock will be the
                //one who releases it.
                dbLocksService
                        .forceUnlock(NonStandardLockCategory.CLUSTER_SINGLETON.getName(), ARTIFACTORY_INIT_LOCK_KEY);
            } else {
                log.warn(
                        "WARNING: Found an init lock in the DB, but the owner is the not the current node. Found: " +
                                "'{}', current: '{}'." +
                                "\n The lock will not be released to allow '{}' to finish the startup phase, and this" +
                                " node" +
                                " will wait until it's released.\n Please note that parallel initialization of nodes " +
                                "during" +
                                " an upgrade or on a vanilla installation is not supported, and you may need to " +
                                "restart" +
                                " this node for it to startup successfully after '{}' finishes booting up.",
                        lockInfo.getOwner(), serverId, lockInfo.getOwner(), lockInfo.getOwner());
            }
        });
    }

    void runInsideInitLockIfNeeded(Runnable task, Supplier<Boolean> shouldAcquireLockToggle) {
        releaseStaleInitLockIfNeeded();
        if (shouldAcquireLockToggle.get()) {
            runInsideLock(task, shouldAcquireLockToggle);
        } else {
            task.run();
        }
    }

    private void releaseStaleInitLockIfNeeded() {
        if (artifactoryHome.isHaConfigured()) {
            File releaseInitLockMarkerFile = artifactoryHome.getReleaseInitLockMarker();
            if (releaseInitLockMarkerFile.exists()) {
                releaseInitLockByMarker(releaseInitLockMarkerFile);
            } else {
                releaseMyInitLock();
            }
        }
    }

    private void runInsideLock(Runnable lockedTask, Supplier<Boolean> shouldAcquireLockToggle) {
        String ownerId = artifactoryHome.getHaNodeProperties().getServerId();
        log.info("Attempting to acquire a high-availability init lock on key {}, with node-id {}",
                ARTIFACTORY_INIT_LOCK_KEY,
                ownerId);
        try {
            dbLocksService.acquireLock(NonStandardLockCategory.CLUSTER_SINGLETON.getName(), ARTIFACTORY_INIT_LOCK_KEY,
                    ownerId, 1,
                    TimeUnit.HOURS);
            log.info("Acquired high-availability init lock on key {}", ARTIFACTORY_INIT_LOCK_KEY);
            if (!shouldAcquireLockToggle.get()) {
                // Some other node my have completed the upgrade or first vanilla init, release lock immediately
                log.info(
                        "Releasing init lock to allow other nodes to continue the init process (upgrade already " +
                                "performed by other node)");
                dbLocksService.unlock(NonStandardLockCategory.CLUSTER_SINGLETON.getName(),
                        ARTIFACTORY_INIT_LOCK_KEY, ownerId);
            }
            lockedTask.run();
        } catch (TimeoutException e) {
            String error = "Unable to acquire high-availability lock on category %s with key %s";
            log.error(error, e);
            throw new IllegalStateException(String.format(error,
                    NonStandardLockCategory.CLUSTER_SINGLETON.getName(), ARTIFACTORY_INIT_LOCK_KEY));
        }
    }

    @SneakyThrows
    private void releaseInitLockByMarker(File releaseInitLockMarkerFile) {
        String releaseInitLockMarkerFileAbsolutePath = releaseInitLockMarkerFile.getAbsolutePath();
        try {
            log.info("Found high-availability release init lock marker file in: {}",
                    releaseInitLockMarkerFileAbsolutePath);
            dbLocksService.forceUnlock(NonStandardLockCategory.CLUSTER_SINGLETON.getName(), ARTIFACTORY_INIT_LOCK_KEY);
        } finally {
            Files.delete(Paths.get(releaseInitLockMarkerFileAbsolutePath));
            log.info("Release high-availability release init lock marker file: {} deleted",
                    releaseInitLockMarkerFileAbsolutePath);
        }
    }

    private Optional<LockInfo> getHaInitLockInfo() {
        return dbLocksService
                .getLockInfoFromDbByCategoryAndLockKey(NonStandardLockCategory.CLUSTER_SINGLETON.getName(),
                        ARTIFACTORY_INIT_LOCK_KEY);
    }
}
