package org.artifactory.storage.db.event.error.dao;

import com.google.common.collect.Lists;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.Pair;
import org.artifactory.common.config.db.ArtifactoryDbProperties;
import org.artifactory.storage.db.event.error.model.NodeEventsErrorsDBO;
import org.artifactory.storage.event.EventType;
import org.jfrog.storage.JdbcHelper;
import org.jfrog.storage.SqlDaoHelper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Collections;
import java.util.List;
import java.util.Optional;
import java.util.stream.Stream;

/**
 * A data access object for the node_events_errors table.
 *
 * @author Mor Merhav
 */
@Repository
public class NodeEventsErrorsDAO {
    private static final String TABLE_NODE_EVENTS_ERRORS = "node_events_errors";
    private static final String[] COLUMNS =
            {"error_id", "first_error_time", "last_error_time", "error_count",
                    "error_message", "operation_key", "event_time",
                    "event_type", "event_category", "path"};
    private static final String PAGINATION_CREATE = TABLE_NODE_EVENTS_ERRORS + " (" + String.join(",", COLUMNS) + ")";

    private final SqlDaoHelper<NodeEventsErrorsDBO> daoHelper;

    @Autowired
    public NodeEventsErrorsDAO(JdbcHelper jdbcHelper, ArtifactoryDbProperties dbProperties) {
        this.daoHelper =
                new SqlDaoHelper<>(jdbcHelper, dbProperties.getDbType(), TABLE_NODE_EVENTS_ERRORS, this::getRowMapper,
                        100);
    }

    public int insert(NodeEventsErrorsDBO dbo) throws SQLException {
        return insert(Collections.singletonList(dbo));
    }

    public int insert(List<NodeEventsErrorsDBO> dbos) throws SQLException {
        return daoHelper.paginationCreate(PAGINATION_CREATE, COLUMNS.length,
                dbo -> Stream.of(dbo.getErrorId(), dbo.getFirstErrorTime(), dbo.getLastErrorTime(), dbo.getErrorCount(),
                        dbo.getErrorMessage(), dbo.getOperationKey(), dbo.getEventTime(), dbo.getEventType(),
                        dbo.getEventCategory(), dbo.getPath()), dbos.toArray(new NodeEventsErrorsDBO[0]));
    }

    public List<NodeEventsErrorsDBO> selectAll(String eventCategory) throws SQLException {
        return Lists.newArrayList(daoHelper.findAllByProperty("event_category", eventCategory));
    }

    public NodeEventsErrorsDBO select(String eventCategory, EventType eventType, String path) throws SQLException {
        return select(eventCategory, eventType, path, null);
    }

    public NodeEventsErrorsDBO select(String eventCategory, EventType eventType, String path, String operationKey)
            throws SQLException {
        String query = "SELECT * FROM " + TABLE_NODE_EVENTS_ERRORS +
                " WHERE event_category = ? AND event_type = ? AND path = ?";
        if (StringUtils.isNotBlank(operationKey)) {
            query = query + " AND operation_key = ?";
            Optional<NodeEventsErrorsDBO> optionalDBO =
                    daoHelper.findFirstByQuery(query, eventCategory, eventType.code(), path, operationKey);
            return optionalDBO.orElse(null);
        }
        Optional<NodeEventsErrorsDBO> optionalDBO =
                daoHelper.findFirstByQuery(query, eventCategory, eventType.code(), path);
        return optionalDBO.orElse(null);
    }

    public List<NodeEventsErrorsDBO> select(String eventCategory, String operationKey) throws SQLException {
        String query = "SELECT * FROM " + TABLE_NODE_EVENTS_ERRORS + " WHERE event_category = ? AND operation_key = ?";
        return daoHelper.findAllByQuery(query, eventCategory, operationKey);
    }

    public boolean delete(String eventCategory, EventType eventType, String path)
            throws SQLException {
        return delete(eventCategory, eventType, path, null);
    }

    public boolean delete(String eventCategory, EventType eventType, String path, String operationKey)
            throws SQLException {
        String query = "DELETE FROM " + TABLE_NODE_EVENTS_ERRORS +
                " WHERE event_category = ? AND event_type = ? AND path = ?";
        if (operationKey != null) {
            query = query + " AND operation_key = ?";
            return daoHelper.executeUpdate(query, eventCategory, eventType.code(), path, operationKey) > 0;
        }
        return daoHelper.executeUpdate(query, eventCategory, eventType.code(), path) > 0;
    }

    public int delete(String eventCategory, String operationKey) throws SQLException {
        return daoHelper.deleteByProperties(
                Lists.newArrayList(Pair.of("event_category", eventCategory), Pair.of("operation_key", operationKey)));
    }

    public int delete(String eventCategory, String operationKey, EventType eventType) throws SQLException {
        return daoHelper.deleteByProperties(
                Lists.newArrayList(Pair.of("event_category", eventCategory), Pair.of("operation_key", operationKey),
                        Pair.of("event_type", eventType.code())));
    }

    /**
     * Updates an existing error (by same event category, path, event type and operation key) record's count, error
     * message and last error time
     *
     * @param dbo - the error record to update
     * @return the number of rows affected
     */
    public int update(NodeEventsErrorsDBO dbo) throws SQLException {
        List<String> findByColumns = Lists.newArrayList("event_category", "event_type", "path");
        List<Object> findByValues = Lists.newArrayList(dbo.getEventCategory(), dbo.getEventType(), dbo.getPath());

        if (dbo.getOperationKey() != null) {
            findByColumns.add("operation_key");
            findByValues.add(dbo.getOperationKey());
        }

        return daoHelper.updateByProperties(
                findByColumns,
                findByValues.stream(),
                List.of("error_count",
                        "error_message",
                        "last_error_time"),
                Stream.of(
                        dbo.getErrorCount(),
                        dbo.getErrorMessage(),
                        dbo.getLastErrorTime()
                )
        );
    }

    private NodeEventsErrorsDBO getRowMapper(ResultSet rs) throws SQLException {
        return NodeEventsErrorsDBO.builder()
                .errorId(rs.getLong("error_id"))
                .eventCategory(rs.getString("event_category"))
                .firstErrorTime(rs.getLong("first_error_time"))
                .lastErrorTime(rs.getLong("last_error_time"))
                .errorCount(rs.getInt("error_count"))
                .errorMessage(rs.getString("error_message"))
                .operationKey(rs.getString("operation_key"))
                .eventTime(rs.getLong("event_time"))
                .eventType(rs.getInt("event_type"))
                .path(rs.getString("path"))
                .build();
    }
}
