/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.webapp.servlet;

import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.security.SecurityService;
import org.artifactory.api.security.UserInfoBuilder;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.security.AuthenticationHelper;
import org.artifactory.security.MutableUserInfo;
import org.artifactory.security.SimpleUser;
import org.artifactory.security.access.AccessTokenAuthentication;
import org.artifactory.security.filters.AuthCacheKey;
import org.artifactory.security.filters.AuthenticationCacheService;
import org.artifactory.security.props.auth.ApiKeyManager;
import org.artifactory.spring.ArtifactoryApplicationContext;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.artifactory.test.MockConfigurationManagerAdapter;
import org.artifactory.test.TestUtils;
import org.artifactory.util.HttpUtils;
import org.artifactory.webapp.servlet.authentication.ArtifactoryBasicAuthenticationEntryPoint;
import org.awaitility.Awaitility;
import org.jfrog.config.ConfigurationManagerImpl;
import org.jfrog.config.ifc.ConfigurationManager;
import org.mockito.Mockito;
import org.springframework.http.MediaType;
import org.springframework.security.authentication.InsufficientAuthenticationException;
import org.springframework.security.core.Authentication;
import org.springframework.security.web.authentication.www.BasicAuthenticationEntryPoint;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import javax.servlet.FilterConfig;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.net.URL;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.TimeUnit;

import static org.easymock.EasyMock.createMock;
import static org.easymock.EasyMock.expect;
import static org.easymock.EasyMock.replay;
import static org.easymock.EasyMock.verify;
import static org.junit.Assert.assertTrue;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertFalse;
import static org.testng.Assert.assertNull;

/**
 * @author Yoav Luft
 */
@Test
public class AccessFilterTest extends ArtifactoryHomeBoundTest {

    public void testAuthenticationChallenge() throws IOException, ServletException {
        ArtifactoryHome artifactoryHome = new ArtifactoryHome("./target/test/testAuthenticationChallenge");
        ConfigurationManager configurationManager = ConfigurationManagerImpl
                .create(new MockConfigurationManagerAdapter(artifactoryHome));
        configurationManager.initDefaultFiles();
        artifactoryHome.initProperties();
        ArtifactoryHome.bind(artifactoryHome);
        ArtifactoryHome.get().getArtifactoryProperties().setProperty(
                ConstantValues.locksTimeoutSecs.getPropertyName(), "10");
        ArtifactoryBasicAuthenticationEntryPoint authenticationEntryPoint =
                new ArtifactoryBasicAuthenticationEntryPoint();
        HttpServletRequest request = createMock(HttpServletRequest.class);

        HttpServletResponse response = createMock(HttpServletResponse.class);
        response.addHeader("WWW-Authenticate", "Basic realm=\"Artifactory Realm\"");
        response.setContentType(MediaType.APPLICATION_JSON_VALUE);
        response.setStatus(401);
        StringWriter out = new StringWriter();
        PrintWriter printWriter = new PrintWriter(out);
        expect(response.getWriter()).andReturn(printWriter);
        expect(request.getRequestURI()).andReturn("testuri");
        expect(request.getHeader("User-Agent")).andStubReturn("the-agent");
        expect(request.getHeader("Request-Agent")).andStubReturn("xx");
        expect(request.getHeader(ApiKeyManager.API_KEY_HEADER)).andStubReturn(null);
        expect(request.getHeader(ApiKeyManager.OLD_API_KEY_HEADER)).andStubReturn(null);
        expect(request.getContextPath()).andReturn("/artifactory");
        replay(request, response);
        authenticationEntryPoint.commence(request, response,
                new InsufficientAuthenticationException("Authentication required"));
        verify(response);
        TestUtils.verifyErrorJson(out.toString(), 401, "Authentication required");
        configurationManager.destroy();
    }

    public void testUiCacheExpiry() throws ServletException {
        AccessFilter accessFilter = new AccessFilter();
        ArtifactoryHome artifactoryHome = ArtifactoryHome.get();
        artifactoryHome.getArtifactoryProperties()
                .setProperty(ConstantValues.securityUiAuthenticationCacheIdleTimeSecs.getPropertyName(), "3");

        FilterConfig filterConfig = Mockito.mock(FilterConfig.class);
        ServletContext servletContext = Mockito.mock(ServletContext.class);
        when(filterConfig.getServletContext()).thenReturn(servletContext);
        when(servletContext.getAttribute(ArtifactoryHome.SERVLET_CTX_ATTR)).thenReturn(artifactoryHome);
        ArtifactoryApplicationContext applicationContext = Mockito.mock(ArtifactoryApplicationContext.class);
        when(servletContext.getAttribute(ArtifactoryContext.APPLICATION_CONTEXT_KEY))
                .thenReturn(applicationContext);
        BasicAuthenticationEntryPoint basicAuthenticationEntryPoint = Mockito.mock(BasicAuthenticationEntryPoint.class);
        when(applicationContext.beanForType(BasicAuthenticationEntryPoint.class))
                .thenReturn(basicAuthenticationEntryPoint);
        SecurityService securityService = Mockito.mock(SecurityService.class);
        when(applicationContext.beanForType(SecurityService.class)).thenReturn(securityService);
        AuthenticationCacheServiceImpl authenticationCacheService = new AuthenticationCacheServiceImpl();
        authenticationCacheService.initialize();
        when(applicationContext.beanForType(AuthenticationCacheService.class)).thenReturn(authenticationCacheService);

        accessFilter.initLater(filterConfig);

        ConcurrentMap<AuthCacheKey, Authentication> uiAuthCache = accessFilter.getUiAuthCache();
        assertTrue(uiAuthCache.isEmpty());

        AuthCacheKey cacheKey = new AuthCacheKey("Authorization: foo", "127.0.0.1");
        uiAuthCache.putIfAbsent(cacheKey,
                Mockito.mock(Authentication.class));
        assertEquals(1, uiAuthCache.size());

        Awaitility.await()
                .atMost(ConstantValues.securityUiAuthenticationCacheIdleTimeSecs.getInt() + 5, TimeUnit.SECONDS)
                .untilAsserted(() -> assertNull(uiAuthCache.get(cacheKey)));
    }

    public void testIsJFrogFrontendAuthentication() {
        String username = "admin";
        MutableUserInfo transientUser = new UserInfoBuilder(username).transientUser().build();
        SimpleUser simpleUser = new SimpleUser(transientUser);
        String frontendTokenValue = "eyJ2ZXIiOiIyIiwidHlwIjoiSldUIiwiYWxnIjoiUlMyNTYiLCJraWQiOiJOR0c3eV9uejl4Z0xYeFhT" +
                "bU5HX3FPMWRpSXVwRzdFOW5VWXltX1V4STVNIn0.eyJzdWIiOiJqZmZlQDAwMCIsInNjcCI6ImFwcGxpZWQtcGVybWlzc2lvbnNc" +
                "L2FkbWluIGFwaToqIiwiYXVkIjoiamZydEAqIiwiaXNzIjoiamZmZUAwMDAiLCJleHAiOjE1NzAxMTU2MTUsImlhdCI6MTU3MDEx" +
                "NTAxNSwianRpIjoiZDc3OGM0MTYtOTEyMi00ODg3LWIwZDAtODFjMWRjZTYzNjY1In0.ZkepryOCH4-RFnbYPxu7Vg3IX0Z-rKyU" +
                "-iYdj17tlYyNnmqKW9drJW_rU8VcGvCW1rYsGKQ9_r9RdRErcT_ckYVKSzT4YJVW5x_GOVriSlvop_Uidi_k_yRu_D6LGwlRa2JU" +
                "jWgDA0Kdd4j2-cP7A6hL7YAMGfswNv-5V2RkyVuiihRkq6nnYG9dRjnW9NtyHmf2yPo7huYtxvkFqF96QkCfa0Ygc90CGLEL5u4x" +
                "aeCTv0iAL_0OkO7n5xu6ewkbJlnySAFo5_9l3YsSTpRmVpkh2_xNOg1_rQVnxnAzARPCAywEcR1Mq1MTKW7kZpz9hBaowxQPW4OO" +
                "AlEgt6UYbA";
        Authentication authentication = new AccessTokenAuthentication(frontendTokenValue, simpleUser, null);

        assertTrue(AuthenticationHelper.isJFrogFrontendAuthentication(authentication));

        String nonFrontendTokenValue = "eyJ2ZXIiOiIyIiwidHlwIjoiSldUIiwiYWxnIjoiUlMyNTYiLCJraWQiOiJheHlVMFFzc3J3S0p3Z" +
                "llqMUtObzNMbzZ6TnhLcnc3NEJ5YkVGVUJrSjB3In0.eyJzdWIiOiJqZnJ0QDAxZGtqMTR5bXNiamczMDRtbWJ4MXExbnNnXC91c" +
                "2Vyc1wvYWxleCIsInNjcCI6Im1lbWJlci1vZi1ncm91cHM6cmVhZGVycyBhcGk6KiIsImF1ZCI6ImpmcnRAMDFka2oxNHltc2JqZ" +
                "zMwNG1tYngxcTFuc2ciLCJpc3MiOiJqZnJ0QDAxZGtqMTR5bXNiamczMDRtbWJ4MXExbnNnXC91c2Vyc1wvYWRtaW4iLCJleHAiO" +
                "jE1NjczMjkyMDAsImlhdCI6MTU2NzMyNTYwMCwianRpIjoiYzBhY2QzYjMtZWI4OS00M2NhLWIxNTMtY2FlZjk5OTMxOWFhIn0.D" +
                "DACJum60qihTypKCZ7irlU1c6AfEmxqFZW5seYwxLsiumbxF3eKiVTHcCaWXNP3fOjsnRKUqie0AerJB7KLlhSSlH2Iu2tFK0tQT" +
                "hTh_Wzt_D27BD1K7ZQAaNLy5p-Leons9jK6HKaDk70wxK3FT08gHaN-HiOslfifbYA_gfOl0EeH5-iW6zOj2sK7d1ixqsNbqbbW4" +
                "VGK1lF1A9YASFWirHgH8Oph0zA02cowyAWGZ6N7NA-oinyldIGhExNkOzQ5HwrWZHyplK6UZwtZ2aXWHihOBwcSjGklzSGqA4Xs1" +
                "n5IBmlLcm__78HxUR0ptOW6U7R6mME0IZ8zJ3-uuA";
        authentication = new AccessTokenAuthentication(nonFrontendTokenValue, simpleUser, null);

        assertFalse(AuthenticationHelper.isJFrogFrontendAuthentication(authentication));
    }

    @Test(dataProvider = "provideGetUrlWithModifiedPort")
    public void testGetUrlWithModifiedPort(String originalHost, int routerPort, String expectedUrl, int artifactoryPort)
            throws IOException {
        AccessFilter accessFilter = new AccessFilter();
        URL urlWithModifiedPort = accessFilter.getPlatformUiUrl(originalHost, routerPort, artifactoryPort);
        assertEquals(urlWithModifiedPort.toExternalForm(), expectedUrl);
    }

    @DataProvider(name = "provideGetUrlWithModifiedPort")
    public static Object[][] provideGetUrlWithModifiedPort() {
        String httpExpectedNoPort = "http://host" + HttpUtils.UNIFIED_JFROG_PLATFORM_UI;
        return new Object[][] {
                {"http://host", 1337, httpExpectedNoPort, 8081},
                {"https://host", 1337, "https://host" + HttpUtils.UNIFIED_JFROG_PLATFORM_UI, 8081},
                {"https://host:8080", 1337, "https://host:8080" + HttpUtils.UNIFIED_JFROG_PLATFORM_UI, 8081},
                {"https://host:8080/", 1337, "https://host:8080" + HttpUtils.UNIFIED_JFROG_PLATFORM_UI, 8081},
                {"http://host:8080/some/path", 1337, "http://host:8080" + HttpUtils.UNIFIED_JFROG_PLATFORM_UI, 8081},
                {"http://host:8081", 1337, "http://host:1337" + HttpUtils.UNIFIED_JFROG_PLATFORM_UI, 8081},
                {"http://host:8081/artifactory", 1337, "http://host:1337" + HttpUtils.UNIFIED_JFROG_PLATFORM_UI, 8081},
                {"http://host/artifactory", 1337, httpExpectedNoPort, 8081},
                {"http://host/some/path", 1337, httpExpectedNoPort, 8081},
                {"http://host/artifactory/webapp", 1337, httpExpectedNoPort, -1},
        };
    }

}
