package org.artifactory.ui.rest.resource.distribution;

import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.release.bundle.ReleaseBundleAddon;
import org.artifactory.addon.release.bundle.airgap.AirGapImportRequestData;
import org.artifactory.addon.release.bundle.airgap.AirGapReleaseBundleException;
import org.artifactory.addon.release.bundle.airgap.AirGapReleaseBundleValidationException;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.rest.constant.SystemRestConstants;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.rest.common.resource.BaseResource;
import org.artifactory.rest.exception.BadRequestException;
import org.artifactory.rest.exception.NotFoundException;
import org.artifactory.ui.rest.service.admin.services.filesystem.BrowseFileSystemService;
import org.artifactory.ui.utils.MultiPartUtils;
import org.artifactory.util.AolUtils;
import org.glassfish.jersey.media.multipart.FormDataParam;
import org.jfrog.common.FileUtils;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import javax.annotation.security.RolesAllowed;
import javax.inject.Inject;
import javax.ws.rs.Consumes;
import javax.ws.rs.DefaultValue;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;

import static org.artifactory.ui.rest.resource.distribution.AirGapReleaseBundleUIResource.PATH_ROOT;

/**
 * @author Yevdo Abramov
 * Created on 23/08/2020
 */
@Path(PATH_ROOT)
@Component
@Scope(BeanDefinition.SCOPE_PROTOTYPE)
@Slf4j
public class AirGapReleaseBundleUIResource extends BaseResource {
    public static final String PATH_ROOT = SystemRestConstants.JFROG_RELEASE;
    private final BrowseFileSystemService browseFileSystemService;
    private final CentralConfigService centralConfigService;
    private final AddonsManager addonsManager;

    @Inject
    public AirGapReleaseBundleUIResource(@NonNull AddonsManager addonsManager,
            @NonNull BrowseFileSystemService browseFileSystemService,
            @NonNull CentralConfigService centralConfigService) {
        this.addonsManager = addonsManager;
        this.browseFileSystemService = browseFileSystemService;
        this.centralConfigService = centralConfigService;
    }

    @Path("/import")
    @RolesAllowed(AuthorizationService.ROLE_ADMIN)
    @Consumes(MediaType.MULTIPART_FORM_DATA)
    @Produces(MediaType.APPLICATION_JSON)
    @POST
    public Response importReleaseBundleFromStream(@QueryParam("dry_run") boolean dryRun,
            @FormDataParam("file") InputStream inputStream) throws AirGapReleaseBundleException {
        AolUtils.assertNotAol("Import release bundle");
        ReleaseBundleAddon releaseBundleAddon = addonsManager.addonByType(ReleaseBundleAddon.class);
        long exportFileSizeInMb =
                FileUtils.bytesToMB(getArtifactoryRestRequest().getServletRequest().getContentLengthLong());
        MultiPartUtils.validateFileSizeLimit(centralConfigService, exportFileSizeInMb);

        try {
            releaseBundleAddon.airGapReleaseBundleImport(new AirGapImportRequestData(inputStream, dryRun));
            return Response.accepted().build();
        } catch (AirGapReleaseBundleValidationException e) {
            throw new BadRequestException(e.getMessage());
        }
    }

    @Path("import/browse")
    @RolesAllowed(AuthorizationService.ROLE_ADMIN)
    @Consumes(MediaType.MULTIPART_FORM_DATA)
    @Produces(MediaType.APPLICATION_JSON)
    @GET
    public Response browseImportDirectory(@QueryParam("path") @DefaultValue("") @NonNull String path)
            throws IOException {
        AolUtils.assertNotAol("Import release bundle");
        File baseImportDir = ContextHelper.get().getArtifactoryHome().getImportDir();
        File importFile = new File(baseImportDir, path);

        // prevent directory traversal attack
        if (!importFile.getCanonicalPath().startsWith(baseImportDir.getCanonicalPath())) {
            throw new BadRequestException(
                    "The path: '" + path + "' must be located under the Artifactory import directory");
        }

        if (!importFile.exists()) {
            throw new NotFoundException("Path: '" + path + "' not found");
        }

        return Response.ok(browseFileSystemService.getFileOrFoldersList(importFile.getPath(), true)).build();
    }
}
