package org.artifactory.rest.resource.release;

import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.io.FileUtils;
import org.apache.http.HttpStatus;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.release.bundle.ReleaseBundleAddon;
import org.artifactory.addon.release.bundle.airgap.AirGapImportRequestData;
import org.artifactory.addon.release.bundle.airgap.AirGapReleaseBundleException;
import org.artifactory.addon.release.bundle.airgap.AirGapReleaseBundleValidationException;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.rest.constant.SystemRestConstants;
import org.artifactory.api.rest.distribution.bundle.models.CloseTransactionStatusResponse;
import org.artifactory.api.rest.release.airgap.AirGapReleaseBundleExportStatusModel;
import org.artifactory.api.rest.release.airgap.AirGapReleaseBundleImportStatusModel;
import org.artifactory.api.rest.release.airgap.AirGapReleaseBundleStatus;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.rest.exception.BadRequestException;
import org.artifactory.rest.exception.NotFoundException;
import org.artifactory.util.AolUtils;
import org.artifactory.util.HttpUtils;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import javax.annotation.security.RolesAllowed;
import javax.inject.Inject;
import javax.servlet.http.HttpServletRequest;
import javax.ws.rs.Consumes;
import javax.ws.rs.DELETE;
import javax.ws.rs.DefaultValue;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;

import static org.artifactory.rest.resource.release.AirGapReleaseBundleResource.PATH_ROOT;

/**
 * @author Yevdo Abramov
 * Created on 20/07/2020
 */
@Slf4j
@Path(PATH_ROOT)
@Component
@Scope(BeanDefinition.SCOPE_PROTOTYPE)
public class AirGapReleaseBundleResource {
    public static final String PATH_ROOT = SystemRestConstants.JFROG_RELEASE;

    private final AddonsManager addonsManager;

    @Context
    private HttpServletRequest servletRequest;

    @Inject
    public AirGapReleaseBundleResource(@NonNull AddonsManager addonsManager) {
        this.addonsManager = addonsManager;
    }

    @Path("/import")
    @RolesAllowed(AuthorizationService.ROLE_ADMIN)
    @Consumes(MediaType.APPLICATION_OCTET_STREAM)
    @Produces(MediaType.APPLICATION_JSON)
    @POST
    public Response importReleaseBundleFromStream(@QueryParam("dry_run") boolean dryRun,
            @NonNull InputStream inputStream) throws AirGapReleaseBundleException {
        AolUtils.assertNotAol("Import release bundle");
        ReleaseBundleAddon releaseBundleAddon = addonsManager.addonByType(ReleaseBundleAddon.class);
        try {
            releaseBundleAddon.airGapReleaseBundleImport(new AirGapImportRequestData(inputStream, dryRun));
            return Response.accepted().build();
        } catch (AirGapReleaseBundleValidationException e) {
            throw new BadRequestException(e.getMessage());
        }
    }

    @Path("/import")
    @RolesAllowed(AuthorizationService.ROLE_ADMIN)
    @Consumes(MediaType.APPLICATION_JSON)
    @Produces(MediaType.APPLICATION_JSON)
    @POST
    public Response importReleaseBundleFromFileSystem(@QueryParam("dry_run") boolean dryRun,
            @QueryParam("path") @DefaultValue("") @NonNull String importFileRelativePath)
            throws AirGapReleaseBundleException, IOException {
        AolUtils.assertNotAol("Import release bundle");
        File baseImportDir = ContextHelper.get().getArtifactoryHome().getImportDir();
        File importFile = new File(baseImportDir, importFileRelativePath);

        // prevent directory traversal attack
        if (!importFile.getCanonicalPath().startsWith(baseImportDir.getCanonicalPath())) {
            throw new BadRequestException(
                    "The file: '" + importFileRelativePath +
                            "' must be located under the Artifactory import directory");
        }

        if (!importFile.exists()) {
            throw new NotFoundException("File: '" + importFileRelativePath + "' not found");
        } else if (!importFile.isFile()) {
            throw new BadRequestException(
                    "Cannot import from: '" + importFileRelativePath + "'. Please specify a proper import file");
        }

        ReleaseBundleAddon releaseBundleAddon = addonsManager.addonByType(ReleaseBundleAddon.class);
        try {
            InputStream importFileInputStream = FileUtils.openInputStream(importFile);
            releaseBundleAddon.airGapReleaseBundleImport(new AirGapImportRequestData(importFileInputStream, dryRun));
            return Response.accepted().build();
        } catch (AirGapReleaseBundleValidationException e) {
            throw new BadRequestException(e.getMessage());
        } catch (IOException e) {
            throw new AirGapReleaseBundleException(e);
        }
    }

    @Path("export/{name}/{version}")
    @Produces(MediaType.APPLICATION_JSON)
    @RolesAllowed(AuthorizationService.ROLE_ADMIN)
    @POST
    public Response exportReleaseBundle(
            @PathParam("name") String bundleName,
            @PathParam("version") String bundleVersion)
            throws AirGapReleaseBundleException {
        AolUtils.assertNotAol("Export release bundle");
        try {
            ReleaseBundleAddon releaseBundleAddon = addonsManager.addonByType(ReleaseBundleAddon.class);
            releaseBundleAddon.triggerAirGapReleaseBundleExport(bundleName, bundleVersion);

            return Response.status(HttpStatus.SC_ACCEPTED).build();
        } catch (AirGapReleaseBundleException e) {
            String errorMessage =
                    String.format("An error occurred while triggering release bundle export for %s/%s: %s", bundleName,
                            bundleVersion, e.getMessage());
            log.error(errorMessage);
            log.debug(errorMessage, e);
            throw e;
        }
    }

    @Path("export/{name}/{version}")
    @Produces(MediaType.APPLICATION_JSON)
    @RolesAllowed(AuthorizationService.ROLE_ADMIN)
    @DELETE
    public Response deleteExportedReleaseBundleFile(
            @PathParam("name") String bundleName,
            @PathParam("version") String bundleVersion) throws AirGapReleaseBundleException {
        AolUtils.assertNotAol("Delete exported release bundle file");
        try {
            ReleaseBundleAddon releaseBundleAddon = addonsManager.addonByType(ReleaseBundleAddon.class);
            releaseBundleAddon.deleteReleaseBundleExportedFile(bundleName, bundleVersion);

            return Response.status(HttpStatus.SC_OK).build();

        } catch (AirGapReleaseBundleException e) {
            String errorMessage =
                    String.format("An error occurred while getting release bundle %s/%s export status: %s", bundleName,
                            bundleVersion, e.getMessage());
            log.error(errorMessage);
            log.debug(errorMessage, e);
            throw e;
        }
    }

    @Path("export/{name}/{version}/status")
    @Produces(MediaType.APPLICATION_JSON)
    @RolesAllowed(AuthorizationService.ROLE_ADMIN)
    @GET
    public Response getExportStatus(@PathParam("name") String bundleName,
            @PathParam("version") String bundleVersion) throws AirGapReleaseBundleException {
        AolUtils.assertNotAol("Export release bundle");
        ReleaseBundleAddon releaseBundleAddon = addonsManager.addonByType(ReleaseBundleAddon.class);
        try {

            String artifactoryUrl = HttpUtils.getServletContextUrl(servletRequest);
            AirGapReleaseBundleExportStatusModel model =
                    releaseBundleAddon.getAirGapExportStatus(bundleName, bundleVersion, artifactoryUrl);

            return Response.ok(model).build();
        } catch (AirGapReleaseBundleException e) {
            String errorMessage =
                    String.format("An error occurred while getting release bundle %s/%s export status: %s", bundleName,
                            bundleVersion, e.getMessage());
            log.error(errorMessage);
            log.debug(errorMessage, e);
            throw e;
        }
    }

    @Path("import/{name}/{version}/status")
    @Produces(MediaType.APPLICATION_JSON)
    @RolesAllowed(AuthorizationService.ROLE_ADMIN)
    @GET
    public Response getImportStatus(@PathParam("name") String bundleName, @PathParam("version") String bundleVersion) {
        AolUtils.assertNotAol("Import release bundle");
        ReleaseBundleAddon releaseBundleAddon = addonsManager.addonByType(ReleaseBundleAddon.class);
        CloseTransactionStatusResponse closeTransactionStatus =
                releaseBundleAddon.checkCloseTransactionStatus(bundleName, bundleVersion);

        AirGapReleaseBundleImportStatusModel importStatus = new AirGapReleaseBundleImportStatusModel();
        switch (closeTransactionStatus.getStatus()) {
            case INPROGRESS:
            case CLOSE_INPROGRESS:
                importStatus.setStatus(AirGapReleaseBundleStatus.IN_PROGRESS);
                break;
            case COMPLETE:
                importStatus.setStatus(AirGapReleaseBundleStatus.COMPLETED);
                break;
            default:
                importStatus.setStatus(AirGapReleaseBundleStatus.FAILED);
                break;
        }

        return Response.ok(importStatus).build();

    }
}
